/*
 * Copyright (c) 2006
 *
 * written by Simon Barrick
 * simon@fourthwall.ndo.co.uk
 *
 * RBFmorpher SOP
 */


//always include these
#include <UT/UT_DSOVersion.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include <PRM/PRM_Include.h>

#include <UT/UT_Interrupt.h>
#include <UT/UT_Vector.h>
#include <UT/UT_Matrix.h>
#include <UT/UT_MatrixSolver.h>


#include "./SOP_RBFmorpher.H"


// This is were the code switches between the standard
// compile and the beta compile.
#ifdef BETA
static const char* const SOP_INTERNAL_NAME = "RBFmorpher_beta";
static const char* const SOP_UI_NAME = "RBF Morpher Beta";
#else
static const char* const SOP_INTERNAL_NAME = "RBFmorpher";
static const char* const SOP_UI_NAME = "RBF Morpher";
#endif


void
newSopOperator( OP_OperatorTable *table )
{
  table->addOperator
  (
    new OP_Operator
    (
      SOP_INTERNAL_NAME,             // Internal name
      SOP_UI_NAME,                   // UI name
      SOP_RBFmorpher::myConstructor,  //
      SOP_RBFmorpher::myTemplateList, //
      3,                             // Min Inputs
      3,                             // Max Inputs
      0                              // Local Variables
      //OP_FLAG_*                    // Additional Flag
    )

// Examples of additional flags include, OP_FLAG_GENERATOR, OP_FLAG_UNORDERED.
// See OP_Operator.h for more details.

  );
}

static PRM_Default  fiveDefault(5);   // Default to 5

static PRM_Name names[] =
{
  PRM_Name("kernel",  "RBF Kernel"),

  PRM_Name( 0 )
};

static PRM_Name kernelMenuNames[] =
{
  PRM_Name( "linear", "Linear" ),

  PRM_Name( "smooth", "Smooth" ),

  PRM_Name( "cube", "Cube" ),

  PRM_Name(0)

};

static PRM_ChoiceList   kernelMenu((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE
                                                    | PRM_CHOICELIST_REPLACE),
                                                      kernelMenuNames);

// How to make choices for a menu.

PRM_Template
SOP_RBFmorpher::myTemplateList[] =
{
  PRM_Template(PRM_ORD, 1, &names[0], PRMzeroDefaults, &kernelMenu),

  PRM_Template(),
};

const char *
SOP_RBFmorpher::inputLabel( unsigned int input) const
{
 switch (input) {
   case 0:
     return "Geometry To Capture";
     break;
   case 1:
     return "Feature Points";
     break;
   case 2:
     return "Target Points";
     break;
   default:
     return "WTF?";
     break;
 }
}

//overriding constructor
OP_Node *
SOP_RBFmorpher::myConstructor( OP_Network *net, const char *name, OP_Operator *op )
{
  return new SOP_RBFmorpher( net, name, op );
}

SOP_RBFmorpher::SOP_RBFmorpher( OP_Network *net, const char *name, OP_Operator *op )
  : SOP_Node( net, name, op )
{
    //mySopFlags.setNeedGuide1( 1 );
}

//destructor
SOP_RBFmorpher::~SOP_RBFmorpher() {}

unsigned
SOP_RBFmorpher::disableParms()
{
    unsigned changed = 0;

    return changed;
}


fpreal64
SOP_RBFmorpher::rbfU(float r)
{
  switch (KERNEL()) {
     case 0:
       return r;
       break;
     case 1:
       return r*r*SYSlog(r*r);
       break;
     case 2:
       return r*r*r;
       break;
     default:
       return 0;
       break;
  }
}
//cook method

OP_ERROR
SOP_RBFmorpher::cookMySop( OP_Context &context )
{
  const GU_Detail       *feature,*target;
  const GEO_Point       *ppt;
  GEO_Point             *opt;
  UT_Interrupt          *boss;
  int                   i,j,k,row,col;


  //check for user interrupts
  boss = UTgetInterrupt();

  boss->opStart("Morphing geometry");

  float now = context.myTime;

  // Before we do anything, we must lock our inputs.  Before returning,
  // we have to make sure that the inputs get unlocked.
  if( lockInputs( context ) >= UT_ERROR_ABORT ) return error();
  if( (feature =   inputGeo(1,context)) == NULL ) return error();
  if( (target =   inputGeo(2,context)) == NULL ) return error();
  if( feature->points().entries() != target->points().entries()) {
    addError(SOP_ERR_INVALID_SRC , ": mismatch point count on inputs 2 and 3");
    return error();
  }

  //copy input geomtry as we aren't creating any
  duplicateSource( 0, context );

//check feature and target point counts are the same

//set up data
  UT_Vector3 p1,p2,p;

  int NfPts = feature->points().entries();
  UT_Vector weightsX(1,NfPts+4);
  UT_Vector weightsY(1,NfPts+4);
  UT_Vector weightsZ(1,NfPts+4);

  UT_Vector targetX(1,NfPts+4);
  UT_Vector targetY(1,NfPts+4);
  UT_Vector targetZ(1,NfPts+4);

  UT_Matrix matrixK(1,NfPts,1,NfPts);
  UT_Matrix matrixQ(1,NfPts,1,4);
  UT_Matrix matrixQTranspose(1,4,1,NfPts);

  UT_Matrix finalMatrix(1,NfPts+4,1,NfPts+4);

//set targetX = x pos of input 2
  targetX.initZero();
  targetY.initZero();
  targetZ.initZero();

  i = 1;
  FOR_ALL_GPOINTS(target,ppt)
  {
    targetX(i) = ppt->getPos()[0];
    targetY(i) = ppt->getPos()[1];
    targetZ(i) = ppt->getPos()[2];
    i ++;
  }

//setup matrixK
  for(row = 1;row <= NfPts; row++)
  {
    for(col = 1;col <= NfPts; col++)
    {
        ppt =  feature->points()(row-1);
        p1 = ppt->getPos();
        ppt =  feature->points()(col-1);
        p2 = ppt->getPos();
        p = p1-p2;

        matrixK(row,col) = rbfU(p.length());
    }
  }

//setup matrixQ
  i = 1;
  FOR_ALL_GPOINTS(feature,ppt)
  {
    matrixQ(i,1) = ppt->getPos()[0];
    matrixQ(i,2) = ppt->getPos()[1];
    matrixQ(i,3) = ppt->getPos()[2];
    matrixQ(i,4) = 1;
    i ++;
  }
//transpose matrixQ
  matrixQ.transpose(matrixQTranspose);

//create finalMatrix
  finalMatrix.initZero();
  for(row=1;row<=NfPts;row++)
    for(col=1;col<=NfPts;col++)
      finalMatrix(row,col)=matrixK(row,col);

  for(row=1;row<=NfPts;row++)
    for(col=NfPts+1;col<=NfPts+4;col++)
      finalMatrix(row,col)=matrixQ(row,col-NfPts);

  for(row=NfPts+1;row<=NfPts+4;row++)
    for(col=1;col<=NfPts;col++)
      finalMatrix(row,col)=matrixQTranspose(row-NfPts,col);

//solve
  UT_MatrixSolver solver;
  float d = 0;
  UT_Permutation P(1,NfPts+4);
  UT_Matrix LU(1,NfPts+4,1,NfPts+4),tLU(1,NfPts+4,1,NfPts+4);

  LU = finalMatrix;
  int test = solver.LUDecomp(LU, P, d);

  if (test == 1)
  {

    weightsX = targetX;
    weightsY = targetY;
    weightsZ = targetZ;
    tLU = LU;
    solver.LUBackSub(LU, P, weightsX);
    LU = tLU;
    solver.LUBackSub(LU, P, weightsY);
    LU = tLU;
    solver.LUBackSub(LU, P, weightsZ);


//cout<<weightsX<<endl;
    //plug the weights back in to deform the input gdp

    FOR_ALL_GPOINTS(gdp,opt)
    {
      fpreal64 outX = weightsX(NfPts+1)*opt->getPos()[0] + weightsX(NfPts+2)*opt->getPos()[1] + weightsX(NfPts+3)*opt->getPos()[2] + weightsX(NfPts+4);
      fpreal64 outY = weightsY(NfPts+1)*opt->getPos()[0] + weightsY(NfPts+2)*opt->getPos()[1] + weightsY(NfPts+3)*opt->getPos()[2] + weightsY(NfPts+4);
      fpreal64 outZ = weightsZ(NfPts+1)*opt->getPos()[0] + weightsZ(NfPts+2)*opt->getPos()[1] + weightsZ(NfPts+3)*opt->getPos()[2] + weightsZ(NfPts+4);

      p1 = opt->getPos();
      for(i=1;i<=NfPts;i++)
      {
        p2 = feature->points()(i-1)->getPos();
        p = p1-p2;
        fpreal64 rF = rbfU(p.length());
        outX += weightsX(i)*rF;
        outY += weightsY(i)*rF;
        outZ += weightsZ(i)*rF;
      }

      UT_Vector3 finalP(outX,outY,outZ);
      opt->getPos() = finalP;
    }
  }

  unlockInputs();

  boss->opEnd();

  return error();

}

/*
OP_ERROR
SOP_RBFmorpher::cookMyGuide1( OP_Context &context )
{
  float now = context.myTime;

  if ( lockInputs( context ) >= UT_ERROR_ABORT ) return error();

  GU_Detail *in_gdp = (GU_Detail *)inputGeo(1, context);
  if (in_gdp)
    myGuide1->copy(*in_gdp);

  unlockInputs();
  return error();
}

*/